/* hexprog.cc   Benutzung des Bootloaders fuer Atmega2560   letzte Aenderung: 3.11.2012 */
#define VERSION "Version 0.5"
/*
History:
29.9.2012	Erstellung aus ser.cc (RP)
8.10.2012       Version 0.4
13.10.2012      Version 0.5
28.10.          Version fuer MacOSX eingefuegt
3.11.           Offensichtlicher Fehler bei memcpy() korrigiert

*/
//#define MACOSX
#define LINUX
//#define ALTESLINUX
//#define CYGWIN

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#ifndef LINUX
#include <sys/ioctl.h>
#endif

/*************************** kleinkram ***************************/
int getline(FILE *fp,char *s,int lim)
{		/* liest eine Textzeile oder maximal lim Zeichen */
		/* und ersetzt den Zeilentrenner durch 0         */
 int c;
 while(--lim && (c=getc(fp))!=EOF && c!='\n')
    if(c!='\r') *s++ = c;
 *s='\0';
 return (c!=EOF);	/* TRUE wenn erfolgreich, FALSE wenn Fileende */
}

int ascitodigit(int c)
{		/* wandelt ein ASCI-Zeichen in eine 1-stellige Hex-Zahl */
 if(!isxdigit(c)) return 0;
 if(islower(c)) c=toupper(c)-'0';
 else c-='0';
 if(c>9) c += '0'-'A'+10;
 return c;
}

/**************** Routinen zur Parameterauswertung ********************/
#define MAXARG 3
static char argflag[128];
void setargflags(char *s)
{
 int c;
 while((c= *s++)!=0)
  {if(c>='a' && c<='z')  c -= 'a'-'A';
   argflag[c&127]=1;
  }
}

bool argflg(char c)
{
 return argflag[(int)c]!=0;
}

/*********************** Serielle Schnittstelle ***********************/
#ifdef MACOSX
//static char sername[80]="/dev/tty.serial-0001"; //fuer aelteren Mac
//static char sername[80]="/dev/tty.usb2uart"; //fuer neuen Mac
static char sername[80]="/dev/tty.SLAB_USBtoUART"; //fuer noch neueren Mac
#endif
#ifdef LINUX
//static char sername[80]="/dev/ttyS0"; //richtiger Seriellport mit Linux
static char sername[80]="/dev/ttyUSB0"; //mit Linux-Kubuntu
#endif
#ifdef ALTESLINUX
//static char sername[80]="/dev/ttyS0"; //richtiger Seriellport mit Linux
static char sername[80]="/dev/ttyUSB0"; //mit Linux-Kubuntu
#endif
#ifdef CYGWIN
static char sername[80]="/dev/com1"; //mit Cygwin unter Windows
#endif

#ifndef BAUD
#define BAUD 500000
#endif
static int baudrate=BAUD;

#ifdef LINUX
//--- Version fuer Linux (Kubuntu11.10) ---
static int fdr,fdw;

static int baudcodliste[]={
   //50,B50, 75,B75, 110,B110, 134,B134, 150,B150,
   300,B300, 600,B600, 1200,B1200, 1800,B1800, 2400,B2400, 4800,B4800,
   9600,B9600, 19200,B19200, 38400,B38400,
   57600,B57600,
   115200,B115200,
   230400,B230400,
   //460800,B460800,
   500000,B500000,
   //576000,B576000,   921600,B921600,
   1000000,B1000000,
   /*
   1152000,B1152000,
   1500000,B1500000,
   2000000,B2000000, 2500000,B2500000, 3000000,B3000000, 3500000,B3500000,
   4000000,B4000000,
   */
   0,0};

int ibaud(int x) //Baudcode in Baudrate umrechnen
{
 int *co,a;
 for(co=baudcodliste;(a= *co++)!=0;co++)
    if(x== *co) return a;
 fprintf(stderr,"Fehler: unbekannter Baudratecode %d\n",x);
 return 0;
}

int baudi(int x) //Baudrate in Baudcode umrechnen
{
 int *co,b,min=0,minb=0;
 for(co=baudcodliste;(b= *co++)>0;co++)
     if(x==b) return *co;
     else if(x>b) {min= *co; minb=b;}
 fprintf(stderr,"Fehler: unbekannte Baudrate %d, auf %d abgerundet\n",x,minb);
 return min;
}

void noecho(int fd)
{
 struct termios buf;
 if(tcgetattr(fd,&buf)<0)
   {printf("tcgetattr() Error\n"); return;}
 buf.c_lflag &= ~ECHO;
 if(baudrate!=0)
     {cfsetispeed(&buf,baudi(baudrate));
      cfsetospeed(&buf,baudi(baudrate));
     }
 if(tcsetattr(fd,TCSAFLUSH,&buf)<0)
   {fprintf(stderr,"tcsetattr() Error\n"); return;}
 int a=cfgetispeed(&buf),b=cfgetospeed(&buf);
 if(argflg('V')) fprintf(stderr,"Baudraten: %d %d\n",ibaud(a),ibaud(b));
 if(a!=b || ibaud(a)!=baudrate)
  fprintf(stderr,"Fehler: konnte Baudrate %d nicht einstellen. Baudrate=%d\n",baudrate,ibaud(a));
}

void seriell_open()
{
 fdr=open(sername,O_RDWR);
 if(fdr<0) {fprintf(stderr,"Fehler: fdr = %d\n",fdr); exit(0);}
	   //wenn dieser Fehler kommt sollte /dev/ttyS0 ueberprueft werden
	   //Zugriffsberechtigung freigeben: chmod a+rw /dev/ttyS0
	   //oder vielleicht /dev/ttyS1
 noecho(fdr);
 fdw=fdr;
}

void seriell_close()
{
 close(fdr);
}

void seriell_readline(char *str,int max)
{
 int i;
 char *s,c;
 for(i=1,s=str;i<max;)
   {read(fdr,&c,1);
    if(c=='\n') break;
    if(c==0)
      {fprintf(stderr,"\\0");//test
      }
    else
      {*s++ = c; i++;}
   }
 if(s!=str && s[-1] == '\r')
   {--s;
    fprintf(stderr,"CR empfangen\n");//test
   }
 *s=0;
}

int seriell_readbyte()
{
 char c;
 read(fdr,&c,1);
 return c;
}

void seriell_writeline(const char *str)
{
 int n=strlen(str);
 char *wbuf=new char[n+3];
 //if(argflag['C']) {sprintf(wbuf,"%s\r\n",str); n++;}
 //else
 sprintf(wbuf,"%s\n",str);
 write(fdw,wbuf,n+1);
 delete wbuf;
}

#else
//--- Version fuer MacOSX und Cygwin und eventuell altes Linux ---
#include <sys/errno.h>

static int fdr,fdw;
static int baudcodliste[]={50,B50, 75,B75, 110,B110, 134,B134, 150,B150,
   300,B300, 600,B600, 1200,B1200, 1800,B1800, 2400,B2400, 4800,B4800,
   9600,B9600, 19200,B19200, 38400,B38400,
   57600,B57600, 115200,B115200,
   230400,B230400,
   //460800,B460800,
   //500000,B500000, //funktioniert auf MacOSX nicht!
   //576000,B576000,
   //921600,B921600, 1000000,B1000000, 1152000,B1152000, 1500000,B1500000,
   //2000000,B2000000, 2500000,B2500000, 3000000,B3000000, 3500000,B3500000,
   //4000000,B4000000,
   0,0};

int ibaud(int x) //Baudcode in Baudrate umrechnen
{
 int *co,a;
 for(co=baudcodliste;(a= *co++)!=0;co++)
    if(x== *co) return a;
 printf("Fehler: unbekannter Baudratencode %d\n",x);
 return 0;
}

int baudi(int x) //Baudrate in Baudcode umrechnen
{
 int *co,b,min=0,minb=0;
 for(co=baudcodliste;(b= *co++)>0;co++)
     if(x==b) return *co;
     else if(x>b) {min= *co; minb=b;}
 fprintf(stderr,"Fehler: unbekannte Baudrate %d, auf %d abgerundet\n",x,minb);
 return min;
}

static struct termios origbuf;
static int origbuf_flag=0;

void noecho(int fd)
{
 int i,a,b;
 struct termios buf;
 if((a=tcgetattr(fd,&buf))<0)
   {printf("tcgetattr() Error %d\n",a); return;}

 //memcpy(&buf,&origbuf,sizeof(struct termios)); //Offensichtlich falsch!
 memcpy(&origbuf,&buf,sizeof(struct termios)); //Es muss von buf nach origbuf kopiert werden.
 origbuf_flag=1; //Flag zum markieren, dass origbuf gueltige Daten enthaelt
 buf.c_cflag = CS8; //8 Bit ohne Parity
 buf.c_cflag |= CREAD;  //Enable input
 buf.c_cflag |= CLOCAL; //Ignore modem cntl
 buf.c_iflag = IGNBRK;
 buf.c_lflag = 0; //no echo input
 for(i=0; i<NCCS; i++)   // REF:JS12052003 Change to clear all option chars
   buf.c_cc[i] = 0;
 buf.c_cc[VMIN] = 0;     // Non blocking input
 buf.c_cc[VTIME] = 0;    // REF:JS12052003 Set to not wait

 if(baudrate!=0)
     {cfsetispeed(&buf,baudi(baudrate));
      cfsetospeed(&buf,baudi(baudrate));
     }

 if((a=tcsetattr(fd,TCSANOW,&buf))<0)
   {printf("tcsetattr() Error %d\n",a); return;}

 a=cfgetispeed(&buf); b=cfgetospeed(&buf);
 fprintf(stderr,"Baudraten: %d %d\n",ibaud(a),ibaud(b)); //test
}

int seriell_open()
{
 fdr=open(sername, O_NONBLOCK | O_RDWR | O_NOCTTY);
 if(fdr<0) return fdr;
 //if(fdr<0) {printf("Fehler: fdr = %d\n",fdr); exit(0);}
	   //wenn dieser Fehler kommt sollte /dev/ttyS0 ueberprueft werden
	   //Zugriffsberechtigung freigeben: chmod a+rw /dev/ttyS0
	   //oder vielleicht /dev/ttyS1 oder /dev/tty.*

#ifdef CYGWIN
 //in Cygwin fehlt TIOCEXCL. Unter Linux ist das in einer 
 //include-Datei als "#define TIOCEXCL 0x540C" gesetzt.
#else
 ioctl(fdr, TIOCEXCL); //set for exclusive OPEN
#endif
 noecho(fdr);
 fdw=fdr;
 return 0;
}

void seriell_close()
{
 if(fdr != -1 && origbuf_flag==1)     //Wenn kein Fehler beim Oeffnen war,
   tcsetattr(fdr, TCSANOW, &origbuf); //dann wieder urspruengliche Einstellungen setzen
 close(fdr);
}

const char* ser_error()
{
 static char buf[16];
 switch(errno)
  {case EBADF: return "EBADF";
   case EFAULT: return "EFAULT";
   case EIO: return "EIO";
   case EINTR: return "EINTR";
   case EINVAL: return "EINVAL";
   case EAGAIN: return "EAGAIN";
   case ENXIO: return "ENXIO";
   case ENOENT: return "ENOENT";
  }
 sprintf(buf,"%d",errno);
 return buf;
}

#ifdef CYGWIN
void ser_read(char *pc)
{
 int i;
 static char buffer[64],*p=NULL;
 static int j=0;
 if(j==0)
   {while((i=read(fdr,buffer,64))==0 || (i<0 && errno==EAGAIN)) ;
    if(i<0) {printf("ser_read() Error: %s\n",ser_error()); return;}
    j=i;
    p=buffer;
   }
 *pc = *p++;
 --j;
}
#else
void ser_read(char *pc)
{
 int i;
 while((i=read(fdr,pc,1))==0 || (i<0 && errno==EAGAIN)) ;
 if(i<0) printf("ser_read() Error: %s\n",ser_error());
}
#endif

int seriell_readc()
{
 char c;
 ser_read(&c);
 return c;
}

void seriell_readline(char *str,int max)
{
 int i;
 char *s,c;
 for(i=1,s=str;i<max;)
   {
    ser_read(&c);
    if(c=='\n') break;
    //if(c==0) printf("\\0");//test
    else
      {*s++ = c; i++;}
   }
 if(s!=str && s[-1] == '\r')
   {--s;
    //printf("CR empfangen\n");//test
   }
 *s=0;
}

void ser_write(char* wbuf,int n)
{
 int i;
 while(n>0)
  {while((i=write(fdw,wbuf,n))==0 || (i<0 && errno==EAGAIN)) ;
   if(i<0) printf("ser_write() Error: %s\n",ser_error());
   if(i<n) wbuf = &wbuf[i];
   n -= i;
  }
}

void seriell_writeline(const char *str)
{
 int n=strlen(str);
 char *wbuf=new char[n+3];
 sprintf(wbuf,"%s\n",str);
 ser_write(wbuf,n+1);
 delete wbuf;
}

#endif
/************************* Hauptprogramm ******************************/
int checksumtest(const char *s);

int main(int argc,char *argv[])
{
 char name[80],name2[80],name3[80],zeile[80],rzeile[80];
 int i,j,c=0;
 FILE *fp=NULL;
 name[0]=name2[0]=name3[0]=0;
 for(j=0,i=1;i<argc;i++)
	{if((c= *argv[i])=='-' || c=='?') setargflags(argv[i]);
	 else	{if(++j==1) strcpy(name2,argv[i]);
	         else if(j==2) strcpy(name3,argv[i]);
	         else if(j==3) strcpy(name,argv[i]);
	}	}
 if(argflg('?') || j>MAXARG)
	{printf("hexprog  %s\n",VERSION);
	 printf("Anwendung: hexprog [/dev/ttyUSB1] [Baudrate] -<Flags> [name.hex]\n");
	 printf("    voreingestellt: %s   %d\n",sername,baudrate);
	 printf(" <Flags>: c=Verbindung pruefen\n");
	 printf("          wf=write flash, Brennen des Programms\n");
	 printf("          we=write EEPROM, Brennen des EEPROMs\n");
	 printf("          rf=read flash, Flash ohne Bootloader-Bereich einlesen\n");
	 printf("          rf2=read flash, inklusive Bootloader einlesen\n");
	 printf("          re=read EEPROM, einlesen und als eeprom. speichern\n");
	 printf("          v=verbose, mit Testausgaben\n");
	 //printf("          x=senden ohne Startzeichen abzuwarten\n");//test
	 printf("Beim Controller muss nach dem Start von hexprog ein Reset ausgeloest werden.\n");
	 printf("Beispiel: EEPROM einlesen und in eeprom.hex speichern:\n");
	 printf("  hexprog /dev/ttyUSB1 -vre >eeprom.hex\n");
	 exit(0);
	}
 if(j>=1)
     {if(isdigit(*name2)) {sscanf(name2,"%d",&baudrate); *name2=0;}
      else if(isdigit(*name3)) {sscanf(name3,"%d",&baudrate); *name3=0;}
      else if(isdigit(*name)) {sscanf(name,"%d",&baudrate); *name=0;}
      if(argflg('V')) fprintf(stderr,"Baudrate = %d\n",baudrate);
      if(name2[0]=='/' || strncmp(name2,"COM",3)==0) {strcpy(sername,name2); *name2=0;}
      else if(name3[0]=='/' || strncmp(name3,"COM",3)==0) {strcpy(sername,name3); *name3=0;}
      else if(name[0]=='/' || strncmp(name,"COM",3)==0) {strcpy(sername,name); *name=0;}
      if(argflg('V')) fprintf(stderr,"SerielleSchnittstelle = \"%s\"\n",sername);
      if(name[0]==0)
       {if(name2[0]!=0) strcpy(name,name2);
	else if(name3[0]!=0) strcpy(name,name3);
       }
      if(argflg('V')) fprintf(stderr,"Name = \"%s\"\n",name);
     }
 if(argflg('V')) fprintf(stderr,"Serielle Schnittstelle '%s' wird geoeffnet\n",sername);
 seriell_open();

 fprintf(stderr,"Jetzt Reset vom Atmega2560 ausloesen. Reset-Knopf oder Spannung Aus/Ein-schalten.\n");
 sprintf(zeile,"%s","StartBootloader"); //Antwort auf Startzeichen vorbereiten

 //if(argflag['X']) {c='X'; sleep(1);} else //test: statt Startbyte empfangen 1 Sekunde warten
 //c=seriell_readbyte(); //Startzeichen empfangen, funktioniert so nicht!

 {seriell_readline(rzeile,80); c=rzeile[0];}  //Startzeichen empfangen
 seriell_writeline(zeile); //und sofort Antwort senden.
 if(argflg('V'))
  {fprintf(stderr,"empfangenes Startzeichen: '%c'\n",c);
   fprintf(stderr,"gesendet: \"%s\"\n",zeile);
  }
 seriell_readline(rzeile,80); if(argflg('V')) fprintf(stderr,"empfangen: \"%s\"\n",rzeile);
 if(argflg('C')) //nur Verbindungstest machen
  {
   printf("Verbindungstest:\n");
   sprintf(zeile,"%s","check");
   seriell_writeline(zeile);    printf("gesendet:  \"%s\"\n",zeile);
   seriell_readline(rzeile,80); printf("empfangen: \"%s\"\n",rzeile);
   sprintf(zeile,"%s","Verbindungstest von hexprog.cc zum Atmega2560");
   seriell_writeline(zeile);    printf("gesendet:  \"%s\"\n",zeile);
   seriell_readline(rzeile,80); printf("empfangen: \"%s\"\n",rzeile);
   sprintf(zeile,"%s","reset");
   seriell_writeline(zeile);    printf("gesendet:  \"%s\"\n",zeile);
   //seriell_readline(rzeile,80); printf("empfangen: \"%s\"\n",rzeile);
  }
 else if(argflg('W') && argflg('F') && name[0]!=0)
  {
   int nfehler=1, nzeilen=0;
   fprintf(stderr,"Datei \"%s\" wird ins Flash gebrannt\n",name);
   fp=fopen(name,"r");
   if(fp==NULL)
    {fprintf(stderr,"Error: cant open \"%s\"\n",name); nfehler=0;}
   else if(strcmp(rzeile,"ok")==0)
    {
     sprintf(zeile,"%s","writeflash");
     if(argflg('V')) fprintf(stderr,"gesendet:  \"%s\"\n",zeile);
     seriell_writeline(zeile);
     seriell_readline(rzeile,80);
     if(strcmp(rzeile,"ok")==0)
      {nfehler=0;
       while(getline(fp,zeile,80))
	{
	 do
	  {if(argflg('V')) fprintf(stderr,"gesendet:  \"%s\"\n",zeile);
	   else if(++nzeilen % 256 == 0) fprintf(stderr,".");//test: nach jeder 256. Zeile einen Punkt
	   seriell_writeline(zeile);
	   seriell_readline(rzeile,80);
	  }
	 while(strcmp(rzeile,"err-checksum")==0 && ++nfehler<5); //Bei Pruefsummenfehler nochmals senden
	 if(strcmp(rzeile,"ok")!=0)
	  {
	   if(strcmp(rzeile,"err-unknownrectyp")==0)
	    {fprintf(stderr,"Warning: unknown rectype ignored: ");
	     fprintf(stderr,"%s\n",zeile);
	    }
	   else break;
	  }
	 nfehler=0;
	}
       if(nfehler==0)
	{
	 seriell_writeline("ok");
	 fprintf(stderr,"\nFlash erfolgreich gebrannt.\n");
	 if(argflg('V')) fprintf(stderr,"gesendet:  \"ok\"\n");
	}
      }
    }
   if(nfehler>0) fprintf(stderr,"Abbruch wegen Fehler: \"%s\"\n",rzeile);
  }
 else if(argflg('R') && argflg('F'))
  {
   fprintf(stderr,"Flash wird eingelesen\n");
   if(strcmp(rzeile,"ok")==0)
    {
     int nzeilen=0;
     if(argflg('2')) sprintf(zeile,"%s","readflash2");
     else sprintf(zeile,"%s","readflash");
     if(argflg('V')) fprintf(stderr,"gesendet:  \"%s\"\n",zeile);
     seriell_writeline(zeile);
     for(;;)
      {int checksum=0;
       seriell_readline(rzeile,80);
       if(rzeile[0]==':' && (checksum=checksumtest(&rzeile[1]))==0)
	{
	 printf("%s\n",rzeile); //Die Hex-Daten auf Standardausgabe oder mit ">name.hex" in die Datei
	 if(++nzeilen % 256 == 0) fprintf(stderr,".");//test: nach jeder 256. Zeile einen Punkt
	 seriell_writeline("ok"); //erfolgreicher Empfang mit "ok" bestaetigen
	}
       else if(strcmp(rzeile,"ok")==0) //Ende der Daten wenn "ok" empfangen wird
	{
	 fprintf(stderr,"\n");//test
	 fprintf(stderr,"Flash erfolgreich eingelesen.\n");
	 break; //erfolgreich fertig
	}
       else if(checksum!=0)
	{
	 fprintf(stderr,"checksum-error - try again\n");
	 seriell_writeline("err-checksum");
	}
       else
	{
	 fprintf(stderr,"Fehler: \"%s\"\n",rzeile); break; //Abbrechen bei unbekanntem Fehler
	}
      }
    }
   else fprintf(stderr,"Abbruch wegen Fehler: \"%s\"\n",rzeile);
  }
 else if(argflg('W') && argflg('E') && name[0]!=0)
  {
   int nfehler=1;
   //if(argflg('V'))
   fprintf(stderr,"Datei \"%s\" wird ins EEPROM gebrannt\n",name);
   fp=fopen(name,"r");
   if(fp==NULL)
    {fprintf(stderr,"Error: cant open \"%s\"\n",name); nfehler=0;}
   else if(strcmp(rzeile,"ok")==0)
    {
     sprintf(zeile,"%s","writeeprom");
     if(argflg('V')) fprintf(stderr,"gesendet:  \"%s\"\n",zeile);
     seriell_writeline(zeile);
     seriell_readline(rzeile,80);
     if(strcmp(rzeile,"ok")==0)
      {nfehler=0;
       while(getline(fp,zeile,80))
	{
	 do
	  {if(argflg('V')) fprintf(stderr,"gesendet:  \"%s\"\n",zeile);
	   seriell_writeline(zeile);
	   seriell_readline(rzeile,80);
	  }
	 while(strcmp(rzeile,"err-checksum")==0 && ++nfehler<5); //Bei Pruefsummenfehler nochmals senden
	 if(strcmp(rzeile,"ok")!=0) break;
	 nfehler=0;
	}
       if(nfehler==0)
	{
	 seriell_writeline("ok");
	 if(argflg('V')) fprintf(stderr,"gesendet:  \"ok\"\n");
	}
      }
    }
   if(nfehler>0) fprintf(stderr,"Abbruch wegen Fehler: \"%s\"\n",rzeile);
  }
 else if(argflg('R') && argflg('E'))
  {
   //if(argflg('V'))
    fprintf(stderr,"EEPROM wird eingelesen\n");
   if(strcmp(rzeile,"ok")==0)
    {
     sprintf(zeile,"%s","readeprom");
     if(argflg('V')) fprintf(stderr,"gesendet:  \"%s\"\n",zeile);
     seriell_writeline(zeile);
     for(;;)
      {int checksum=0;
       seriell_readline(rzeile,80);
       if(rzeile[0]==':' && (checksum=checksumtest(&rzeile[1]))==0)
	{
	 printf("%s\n",rzeile); //Die Hex-Daten auf Standardausgabe oder mit ">name.hex" in die Datei
	 seriell_writeline("ok"); //erfolgreicher Empfang mit "ok" bestaetigen
	}
       else if(strcmp(rzeile,"ok")==0) //Ende der Daten wenn "ok" empfangen wird
	{
	 fprintf(stderr,"EEPROM erfolgreich eingelesen.\n");
	 break; //erfolgreich fertig
	}
       else if(checksum!=0)
	{
	 fprintf(stderr,"checksum-error - try again\n");
	 seriell_writeline("err-checksum");
	}
       else
	{
	 fprintf(stderr,"Fehler: \"%s\"\n",rzeile); break; //Abbrechen bei unbekanntem Fehler
	}
      }
    }
   else fprintf(stderr,"Abbruch wegen Fehler: \"%s\"\n",rzeile);
  }
 else if(argflg('W') && (argflg('F') || argflg('E')) && name[0]==0)
  {
   fprintf(stderr,"Fehlender Dateiname.hex zum Daten brennen. Nichts gemacht.\n");
  }
 else
  {
   fprintf(stderr,"Fehlendes oder falsches Kommando. Nichts gemacht.\n");
  }

 if(argflg('V')) fprintf(stderr,"Serielle Schnittstelle wird geschlossen\n");
 seriell_close();
 return 0;
}/* ende von main */

int checksumtest(const char *s)
{
 int pruefsumme=0,c1,c2;
 while((c1= *s++)!=0 && (c2= *s++)!=0)
  {
   pruefsumme += (ascitodigit(c1)<<4)+ascitodigit(c2);
  }
 return (pruefsumme&0xFF);
}
