/* hexprog.cc   Benutzung des Bootloaders fuer Atmega2560   letzte Aenderung: 28.10.2012 */
#define VERSION "Version 0.5m"
/*
History:
29.9.2012	Erstellung aus ser.cc (RP)
8.10.2012       Version 0.4
13.10.12        Version fuer MacOSX, geht noch nicht mit hoher Baudrate

*/
#define MACOSX
//#define LINUX
//#define CYGWIN

#define NONBLOCKMODUS //kann unter Linux auskommentiert werden

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <sys/ioctl.h>

/*************************** kleinkram ***************************/
int getline(FILE *fp,char *s,int lim)
{		/* liest eine Textzeile oder maximal lim Zeichen */
		/* und ersetzt den Zeilentrenner durch 0         */
 int c;
 while(--lim && (c=getc(fp))!=EOF && c!='\n')
    if(c!='\r') *s++ = c;
 *s='\0';
 return (c!=EOF);	/* TRUE wenn erfolgreich, FALSE wenn Fileende */
}

int ascitodigit(int c)
{		/* wandelt ein ASCI-Zeichen in eine 1-stellige Hex-Zahl */
 if(!isxdigit(c)) return 0;
 if(islower(c)) c=toupper(c)-'0';
 else c-='0';
 if(c>9) c += '0'-'A'+10;
 return c;
}

/**************** Routinen zur Parameterauswertung ********************/
#define MAXARG 3
static char argflag[128];
void setargflags(char *s)
{
 int c;
 while((c= *s++)!=0)
  {if(c>='a' && c<='z')  c -= 'a'-'A';
   argflag[c&127]=1;
  }
}

bool argflg(char c)
{
 return argflag[(int)c]!=0;
}

/*********************** Serielle Schnittstelle ***********************/
#ifdef MACOSX
//static char sername[80]="/dev/tty.serial-0001"; //fuer aelteren Mac
//static char sername[80]="/dev/tty.usb2uart"; //fuer neuen Mac
static char sername[80]="/dev/tty.SLAB_USBtoUART"; //fuer noch neueren Mac
#endif
#ifdef LINUX
//static char sername[80]="/dev/ttyS0"; //richtiger Seriellport mit Linux
static char sername[80]="/dev/ttyUSB0"; //mit Linux-Kubuntu
#endif
#ifdef CYGWIN
static char sername[80]="/dev/com1"; //mit Cygwin unter Windows
#endif

#ifndef BAUD
#define BAUD 115200
#endif
static int baudrate=BAUD;

#include "ser_macosx.cc"

/************************* Hauptprogramm ******************************/
int checksumtest(const char *s);

int main(int argc,char *argv[])
{
 char name[80],name2[80],name3[80],zeile[80],rzeile[80];
 int i,j,c=0;
 FILE *fp=NULL;
 name[0]=name2[0]=name3[0]=0;
 for(j=0,i=1;i<argc;i++)
	{if((c= *argv[i])=='-' || c=='?') setargflags(argv[i]);
	 else	{if(++j==1) strcpy(name2,argv[i]);
	         else if(j==2) strcpy(name3,argv[i]);
	         else if(j==3) strcpy(name,argv[i]);
	}	}
 if(argflg('?') || j>MAXARG)
	{printf("hexprog  %s\n",VERSION);
	 printf("Anwendung: hexprog [/dev/ttyUSB1] [Baudrate] -<Flags> [name.hex]\n");
	 printf("    voreingestellt: %s   %d\n",sername,baudrate);
	 printf(" <Flags>: c=Verbindung pruefen\n");
	 printf("          wf=write flash, Brennen des Programms\n");
	 printf("          we=write EEPROM, Brennen des EEPROMs\n");
	 printf("          rf=read flash, Flash ohne Bootloader-Bereich einlesen\n");
	 printf("          rf2=read flash, inklusive Bootloader einlesen\n");
	 printf("          re=read EEPROM, einlesen und als eeprom. speichern\n");
	 printf("          v=verbose, mit Testausgaben\n");
	 //printf("          x=senden ohne Startzeichen abzuwarten\n");//test
	 printf("Beim Controller muss nach dem Start von hexprog ein Reset ausgeloest werden.\n");
	 printf("Beispiel: EEPROM einlesen und in eeprom.hex speichern:\n");
	 printf("  hexprog /dev/ttyUSB1 -vre >eeprom.hex\n");
	 exit(0);
	}
 if(j>=1)
     {if(isdigit(*name2)) {sscanf(name2,"%d",&baudrate); *name2=0;}
      else if(isdigit(*name3)) {sscanf(name3,"%d",&baudrate); *name3=0;}
      else if(isdigit(*name)) {sscanf(name,"%d",&baudrate); *name=0;}
      if(argflg('V')) fprintf(stderr,"Baudrate = %d\n",baudrate);
      if(name2[0]=='/' || strncmp(name2,"COM",3)==0) {strcpy(sername,name2); *name2=0;}
      else if(name3[0]=='/' || strncmp(name3,"COM",3)==0) {strcpy(sername,name3); *name3=0;}
      else if(name[0]=='/' || strncmp(name,"COM",3)==0) {strcpy(sername,name); *name=0;}
      if(argflg('V')) fprintf(stderr,"SerielleSchnittstelle = \"%s\"\n",sername);
      if(name[0]==0)
       {if(name2[0]!=0) strcpy(name,name2);
	else if(name3[0]!=0) strcpy(name,name3);
       }
      if(argflg('V')) fprintf(stderr,"Name = \"%s\"\n",name);
     }
 if(argflg('V')) fprintf(stderr,"Serielle Schnittstelle '%s' wird geoeffnet\n",sername);
 seriell_open();

 fprintf(stderr,"Jetzt Reset vom Atmega2560 ausloesen. Reset-Knopf oder Spannung Aus/Ein-schalten.\n");
 sprintf(zeile,"%s","StartBootloader"); //Antwort auf Startzeichen vorbereiten

 //if(argflag['X']) {c='X'; sleep(1);} else //test: statt Startbyte empfangen 1 Sekunde warten
 //c=seriell_readbyte(); //Startzeichen empfangen, funktioniert so nicht!

 {seriell_readline(rzeile,80); c=rzeile[0];}  //Startzeichen empfangen
 seriell_writeline(zeile); //und sofort Antwort senden.
 if(argflg('V'))
  {fprintf(stderr,"empfangenes Startzeichen: '%c'\n",c);
   fprintf(stderr,"gesendet: \"%s\"\n",zeile);
  }
 seriell_readline(rzeile,80); if(argflg('V')) fprintf(stderr,"empfangen: \"%s\"\n",rzeile);
 if(argflg('C')) //nur Verbindungstest machen
  {
   printf("Verbindungstest:\n");
   sprintf(zeile,"%s","check");
   seriell_writeline(zeile);    printf("gesendet:  \"%s\"\n",zeile);
   seriell_readline(rzeile,80); printf("empfangen: \"%s\"\n",rzeile);
   sprintf(zeile,"%s","Verbindungstest von hexprog.cc zum Atmega2560");
   seriell_writeline(zeile);    printf("gesendet:  \"%s\"\n",zeile);
   seriell_readline(rzeile,80); printf("empfangen: \"%s\"\n",rzeile);
   sprintf(zeile,"%s","reset");
   seriell_writeline(zeile);    printf("gesendet:  \"%s\"\n",zeile);
   //seriell_readline(rzeile,80); printf("empfangen: \"%s\"\n",rzeile);
  }
 else if(argflg('W') && argflg('F') && name[0]!=0)
  {
   int nfehler=1, nzeilen=0;
   fprintf(stderr,"Datei \"%s\" wird ins Flash gebrannt\n",name);
   fp=fopen(name,"r");
   if(fp==NULL)
    {fprintf(stderr,"Error: cant open \"%s\"\n",name); nfehler=0;}
   else if(strcmp(rzeile,"ok")==0)
    {
     sprintf(zeile,"%s","writeflash");
     if(argflg('V')) fprintf(stderr,"gesendet:  \"%s\"\n",zeile);
     seriell_writeline(zeile);
     seriell_readline(rzeile,80);
     if(strcmp(rzeile,"ok")==0)
      {nfehler=0;
       while(getline(fp,zeile,80))
	{
	 do
	  {if(argflg('V')) fprintf(stderr,"gesendet:  \"%s\"\n",zeile);
	   else if(++nzeilen % 256 == 0) fprintf(stderr,".");//test: nach jeder 256. Zeile einen Punkt
	   seriell_writeline(zeile);
	   seriell_readline(rzeile,80);
	  }
	 while(strcmp(rzeile,"err-checksum")==0 && ++nfehler<5); //Bei Pruefsummenfehler nochmals senden
	 if(strcmp(rzeile,"ok")!=0)
	  {
	   if(strcmp(rzeile,"err-unknownrectyp")==0)
	    {fprintf(stderr,"Warning: unknown rectype ignored: ");
	     fprintf(stderr,"%s\n",zeile);
	    }
	   else break;
	  }
	 nfehler=0;
	}
       if(nfehler==0)
	{
	 seriell_writeline("ok");
	 fprintf(stderr,"\nFlash erfolgreich gebrannt.\n");
	 if(argflg('V')) fprintf(stderr,"gesendet:  \"ok\"\n");
	}
      }
    }
   if(nfehler>0) fprintf(stderr,"Abbruch wegen Fehler: \"%s\"\n",rzeile);
  }
 else if(argflg('R') && argflg('F'))
  {
   fprintf(stderr,"Flash wird eingelesen\n");
   if(strcmp(rzeile,"ok")==0)
    {
     int nzeilen=0;
     if(argflg('2')) sprintf(zeile,"%s","readflash2");
     else sprintf(zeile,"%s","readflash");
     if(argflg('V')) fprintf(stderr,"gesendet:  \"%s\"\n",zeile);
     seriell_writeline(zeile);
     for(;;)
      {int checksum=0;
       seriell_readline(rzeile,80);
       if(rzeile[0]==':' && (checksum=checksumtest(&rzeile[1]))==0)
	{
	 printf("%s\n",rzeile); //Die Hex-Daten auf Standardausgabe oder mit ">name.hex" in die Datei
	 if(++nzeilen % 256 == 0) fprintf(stderr,".");//test: nach jeder 256. Zeile einen Punkt
	 seriell_writeline("ok"); //erfolgreicher Empfang mit "ok" bestaetigen
	}
       else if(strcmp(rzeile,"ok")==0) //Ende der Daten wenn "ok" empfangen wird
	{
	 fprintf(stderr,"\n");//test
	 fprintf(stderr,"Flash erfolgreich eingelesen.\n");
	 break; //erfolgreich fertig
	}
       else if(checksum!=0)
	{
	 fprintf(stderr,"checksum-error - try again\n");
	 seriell_writeline("err-checksum");
	}
       else
	{
	 fprintf(stderr,"Fehler: \"%s\"\n",rzeile); break; //Abbrechen bei unbekanntem Fehler
	}
      }
    }
   else fprintf(stderr,"Abbruch wegen Fehler: \"%s\"\n",rzeile);
  }
 else if(argflg('W') && argflg('E') && name[0]!=0)
  {
   int nfehler=1;
   //if(argflg('V'))
   fprintf(stderr,"Datei \"%s\" wird ins EEPROM gebrannt\n",name);
   fp=fopen(name,"r");
   if(fp==NULL)
    {fprintf(stderr,"Error: cant open \"%s\"\n",name); nfehler=0;}
   else if(strcmp(rzeile,"ok")==0)
    {
     sprintf(zeile,"%s","writeeprom");
     if(argflg('V')) fprintf(stderr,"gesendet:  \"%s\"\n",zeile);
     seriell_writeline(zeile);
     seriell_readline(rzeile,80);
     if(strcmp(rzeile,"ok")==0)
      {nfehler=0;
       while(getline(fp,zeile,80))
	{
	 do
	  {if(argflg('V')) fprintf(stderr,"gesendet:  \"%s\"\n",zeile);
	   seriell_writeline(zeile);
	   seriell_readline(rzeile,80);
	  }
	 while(strcmp(rzeile,"err-checksum")==0 && ++nfehler<5); //Bei Pruefsummenfehler nochmals senden
	 if(strcmp(rzeile,"ok")!=0) break;
	 nfehler=0;
	}
       if(nfehler==0)
	{
	 seriell_writeline("ok");
	 if(argflg('V')) fprintf(stderr,"gesendet:  \"ok\"\n");
	}
      }
    }
   if(nfehler>0) fprintf(stderr,"Abbruch wegen Fehler: \"%s\"\n",rzeile);
  }
 else if(argflg('R') && argflg('E'))
  {
   //if(argflg('V'))
    fprintf(stderr,"EEPROM wird eingelesen\n");
   if(strcmp(rzeile,"ok")==0)
    {
     sprintf(zeile,"%s","readeprom");
     if(argflg('V')) fprintf(stderr,"gesendet:  \"%s\"\n",zeile);
     seriell_writeline(zeile);
     for(;;)
      {int checksum=0;
       seriell_readline(rzeile,80);
       if(rzeile[0]==':' && (checksum=checksumtest(&rzeile[1]))==0)
	{
	 printf("%s\n",rzeile); //Die Hex-Daten auf Standardausgabe oder mit ">name.hex" in die Datei
	 seriell_writeline("ok"); //erfolgreicher Empfang mit "ok" bestaetigen
	}
       else if(strcmp(rzeile,"ok")==0) //Ende der Daten wenn "ok" empfangen wird
	{
	 fprintf(stderr,"EEPROM erfolgreich eingelesen.\n");
	 break; //erfolgreich fertig
	}
       else if(checksum!=0)
	{
	 fprintf(stderr,"checksum-error - try again\n");
	 seriell_writeline("err-checksum");
	}
       else
	{
	 fprintf(stderr,"Fehler: \"%s\"\n",rzeile); break; //Abbrechen bei unbekanntem Fehler
	}
      }
    }
   else fprintf(stderr,"Abbruch wegen Fehler: \"%s\"\n",rzeile);
  }
 else if(argflg('W') && (argflg('F') || argflg('E')) && name[0]==0)
  {
   fprintf(stderr,"Fehlender Dateiname.hex zum Daten brennen. Nichts gemacht.\n");
  }
 else
  {
   fprintf(stderr,"Fehlendes oder falsches Kommando. Nichts gemacht.\n");
  }

 if(argflg('V')) fprintf(stderr,"Serielle Schnittstelle wird geschlossen\n");
 seriell_close();
 return 0;
}/* ende von main */

int checksumtest(const char *s)
{
 int pruefsumme=0,c1,c2;
 while((c1= *s++)!=0 && (c2= *s++)!=0)
  {
   pruefsumme += (ascitodigit(c1)<<4)+ascitodigit(c2);
  }
 return (pruefsumme&0xFF);
}
