/* ser2.cc  Unterprogramme fuer Serielle Schnittstelle
History:
11.10.2012	Erstellung aus ser.cc und rflash.cc

*/

//nur eine der 3 folgenden Zeilen nicht auskommentiert:
//#define LINUX
#define MACOSX
//#define CYGWIN 1 //Version fuer Windows

#include <sys/errno.h>

//static int baudrate=38400; //im Hauptprogramm definieren

static int fdr,fdw;
static int baudcodliste[]={50,B50, 75,B75, 110,B110, 134,B134, 150,B150,
   300,B300, 600,B600, 1200,B1200, 1800,B1800, 2400,B2400, 4800,B4800,
   9600,B9600, 19200,B19200, 38400,B38400,
   57600,B57600, 115200,B115200,
   230400,B230400,
   //460800,B460800,
   500000,B500000, //funktioniert auf MacOSX nicht!
   //576000,B576000,
   //921600,B921600, 1000000,B1000000, 1152000,B1152000, 1500000,B1500000,
   //2000000,B2000000, 2500000,B2500000, 3000000,B3000000, 3500000,B3500000,
   //4000000,B4000000,
   0,0};

int ibaud(int x) //Baudcode in Baudrate umrechnen
{
 int *co,a;
 for(co=baudcodliste;(a= *co++)!=0;co++)
    if(x== *co) return a;
 printf("Fehler: unbekannter Baudratencode %d\n",x);
 return 0;
}

int baudi(int x) //Baudrate in Baudcode umrechnen
{
 int *co,b,min=0;
 for(co=baudcodliste;(b= *co++)>0;co++)
     if(x==b) return *co;
     else if(x>b) {min= *co;}
 printf("Warnung: unbekannte Baudrate %d, auf %d abgerundet\n",x,min);
 return min;
}

static struct termios origbuf;

void noecho(int fd)
{
 int i,a,b;
 struct termios buf;
 if((a=tcgetattr(fd,&buf))<0)
   {printf("tcgetattr() Error %d\n",a); return;}

 memcpy(&buf,&origbuf,sizeof(struct termios));
 buf.c_cflag = CS8; //8 Bit ohne Parity
 buf.c_cflag |= CREAD;  //Enable input
 buf.c_cflag |= CLOCAL; //Ignore modem cntl
 buf.c_iflag = IGNBRK;
 buf.c_lflag = 0; //no echo input
 for(i=0; i<NCCS; i++)   // REF:JS12052003 Change to clear all option chars
   buf.c_cc[i] = 0;
 buf.c_cc[VMIN] = 0;     // Non blocking input
 buf.c_cc[VTIME] = 0;    // REF:JS12052003 Set to not wait

 if(baudrate!=0)
     {cfsetispeed(&buf,baudi(baudrate));
      cfsetospeed(&buf,baudi(baudrate));
     }

 if((a=tcsetattr(fd,TCSANOW,&buf))<0)
   {printf("tcsetattr() Error %d\n",a); return;}

 a=cfgetispeed(&buf); b=cfgetospeed(&buf);
 fprintf(stderr,"Baudraten: %d %d\n",ibaud(a),ibaud(b)); //test
}

int seriell_open()
{
 fdr=open(sername, O_NONBLOCK | O_RDWR | O_NOCTTY);
 if(fdr<0) return fdr;
 //if(fdr<0) {printf("Fehler: fdr = %d\n",fdr); exit(0);}
	   //wenn dieser Fehler kommt sollte /dev/ttyS0 ueberprueft werden
	   //Zugriffsberechtigung freigeben: chmod a+rw /dev/ttyS0
	   //oder vielleicht /dev/ttyS1 oder /dev/tty.*

#ifdef CYGWIN
 //in Cygwin fehlt TIOCEXCL. Unter Linux ist das in einer 
 //include-Datei als "#define TIOCEXCL 0x540C" gesetzt.
#else
 ioctl(fdr, TIOCEXCL); //set for exclusive OPEN
#endif
 noecho(fdr);
 fdw=fdr;
 return 0;
}

void seriell_close()
{
 if(fdr != -1) tcsetattr(fdr, TCSANOW, &origbuf);
 close(fdr);
}

const char* ser_error()
{
 static char buf[16];
 switch(errno)
  {case EBADF: return "EBADF";
   case EFAULT: return "EFAULT";
   case EIO: return "EIO";
   case EINTR: return "EINTR";
   case EINVAL: return "EINVAL";
   case EAGAIN: return "EAGAIN";
   case ENXIO: return "ENXIO";
   case ENOENT: return "ENOENT";
  }
 sprintf(buf,"%d",errno);
 return buf;
}

#ifdef CYGWIN
void ser_read(char *pc)
{
 int i;
 static char buffer[64],*p=NULL;
 static int j=0;
 if(j==0)
   {while((i=read(fdr,buffer,64))==0 || (i<0 && errno==EAGAIN)) ;
    if(i<0) {printf("ser_read() Error: %s\n",ser_error()); return;}
    j=i;
    p=buffer;
   }
 *pc = *p++;
 --j;
}
#else
void ser_read(char *pc)
{
 int i;
 while((i=read(fdr,pc,1))==0 || (i<0 && errno==EAGAIN)) ;
 if(i<0) printf("ser_read() Error: %s\n",ser_error());
}
#endif

int seriell_readc()
{
 char c;
 ser_read(&c);
 return c;
}

void seriell_readline(char *str,int max)
{
 int i;
 char *s,c;
 for(i=1,s=str;i<max;)
   {
    ser_read(&c);
    if(c=='\n') break;
    //if(c==0) printf("\\0");//test
    else
      {*s++ = c; i++;}
   }
 if(s!=str && s[-1] == '\r')
   {--s;
    //printf("CR empfangen\n");//test
   }
 *s=0;
}

void ser_write(char* wbuf,int n)
{
 int i;
 while(n>0)
  {while((i=write(fdw,wbuf,n))==0 || (i<0 && errno==EAGAIN)) ;
   if(i<0) printf("ser_write() Error: %s\n",ser_error());
   if(i<n) wbuf = &wbuf[i];
   n -= i;
  }
}

void seriell_writeline(const char *str)
{
 int n=strlen(str);
 char *wbuf=new char[n+3];
 sprintf(wbuf,"%s\n",str);
 ser_write(wbuf,n+1);
 delete wbuf;
}

int ser_verbindungstest()
{
 char zeile[40]; zeile[0]=0;
 seriell_writeline("v"); //Version vom Scatter-Geraet anfordern
 seriell_readline(zeile,40);
 //if(argflag['V'])
  printf("%s\n",zeile);
 if(*zeile!=':')
   {seriell_close();
    return -1; //Fehler
   }
 return 0; //ok
}
