// uart_test.c
#include <avr/io.h>
#include <inttypes.h>
#ifndef bool
#define bool int8_t
#endif

#define F_CPU 16000000UL
#define BAUD 500000UL  // Baudrate

// Berechnungen:
#define UBRR_VAL ((F_CPU+BAUD*8)/(BAUD*16)-1)  // clever runden
#define BAUD_REAL (F_CPU/(16*(UBRR_VAL+1)))    // Reale Baudrate
#define BAUD_ERROR ((BAUD_REAL*1000)/BAUD)     // Fehler in Promille, 1000 = kein Fehler.
#if ((BAUD_ERROR<990) || (BAUD_ERROR>1010))
  #error Systematischer Fehler der Baudrate groesser 1% und damit zu hoch! 
#endif

void uart_init()
{
 UBRR3 = UBRR_VAL;
 UCSR3B = (1<<RXEN3) | (1<<TXEN3); // UART3 RX und TX einschalten
 //UCSR3C = (1<<USBS3)|(3<<UCSZ30); // Asynchron 8N2
 UCSR3C = (3<<UCSZ30); // Asynchron 8N1
}

void uart_putc(char c)
{
 while (!(UCSR3A & (1<<UDRE3))) ;// warten bis Senden moeglich
 UDR3 = c;                       // sende Zeichen
}

bool uart_getcheck()
{
 return (UCSR3A & (1<<RXC3)); //true wenn ein Zeichen verfuegbar
}

char uart_getc(void)
{
 while (!(UCSR3A & (1<<RXC3))) ;// warten bis Zeichen verfuegbar
 return UDR3; // Zeichen aus UDR an Aufrufer zurueckgeben
}

void uart_write(const char *s)
{
 char c;
 while((c= *s++)!=0) uart_putc(c);
}

void uart_getzeile(char *zeile,int max)
{
 char c;
 while(1)
  {if(uart_getcheck())
    {
     c = uart_getc(); if(c=='\n' || max<=1) {*zeile=0; return;}
     if(c>=' ') {*zeile++ = c; --max;}
    }
  }
}

int main()
{
 char zeile[80];
 uart_init();
 uart_write("Hallo Welt\n"); //wenn Gegenstelle ein Windows-Programm ist, eventuell "\r\n" machen.
 while(1)
  {
   uart_getzeile(zeile,80);
   uart_write(zeile); uart_write("-ok\n");
  }
 return 0;
}
